#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

from lightgbm import LGBMRegressor as _LGBMRegressor
from sklearn.model_selection import train_test_split

from amplo.classification.lgbm import _validate_lightgbm_callbacks
from amplo.regression._base import BaseRegressor
from amplo.utils import check_dtypes


class LGBMRegressor(BaseRegressor):
    """
    Amplo wrapper for lightgbm.LGBMRegressor.

    Parameters
    ----------
    callbacks : list of str, optional
        The following callbacks are currently supported:
            - early stopping, "early_stopping_rounds=100"
    test_size : float, default: 0.1
        Test size for train-test-split in fitting the model.
    random_state : int, default: None
        Random state for train-test-split in fitting the model.
    verbose : {0, 1, 2}, default: 0
        Verbose logging.
    **model_params : Any
        Model parameters for underlying lightgbm.LGBMRegressor.
    """

    model: _LGBMRegressor  # type hint

    def __init__(
        self,
        callbacks: list[str] | None = None,
        test_size: float = 0.1,
        random_state: int | None = None,
        verbose: int = 0,
        **model_params,
    ):
        # Verify input dtypes and integrity
        check_dtypes(
            ("callbacks", callbacks, (type(None), list)),
            ("test_size", test_size, float),
            ("random_state", random_state, (type(None), int)),
            ("model_params", model_params, dict),
        )
        if not 0 <= test_size < 1:
            raise ValueError(f"Invalid attribute for test_size: {test_size}")

        # Set up callbacks
        callbacks = callbacks or []
        for cb_name, cb_default_value in [("early_stopping_rounds", 100)]:
            # Skip if already present in callbacks
            if any(callback.startswith(cb_name) for callback in callbacks):
                continue
            # Pop model parameters into callbacks
            callbacks.append(f"{cb_name}={model_params.pop(cb_name, cb_default_value)}")

        # Set up model parameters
        default_model_params = {
            "n_estimators": 1000,  # number of boosting rounds
            "force_col_wise": True,  # reduce memory cost
            "verbosity": verbose - 1,  # don't use "verbose" due to self.reset()
        }
        for k, v in default_model_params.items():
            if k not in model_params:
                model_params[k] = v
        model = _LGBMRegressor(**model_params)

        # Set attributes
        self.callbacks = callbacks
        self.test_size = test_size
        self.random_state = random_state

        super().__init__(model=model, verbose=verbose)

    def _fit(self, x, y=None, **fit_params):
        # Set up fitting callbacks
        callbacks = _validate_lightgbm_callbacks(self.callbacks)

        # Split data and fit model
        xt, xv, yt, yv = train_test_split(
            x, y, test_size=self.test_size, random_state=self.random_state
        )
        self.model.fit(
            xt, yt, eval_set=[(xv, yv)], callbacks=callbacks, eval_metric="rmse"
        )



#  Copyright (c) 2022 by Amplo.

import json
import warnings

import numpy as np
import pytest
import scipy.stats

from amplo.automl import DriftDetector
from amplo.automl.drift_detection import DataDriftWarning
from amplo.utils.testing import (
    DummyDataSampler,
    make_cat_data,
    make_data,
    make_num_data,
)


class DummyPredictor(DummyDataSampler):
    def predict(self, data):
        return self.sample_data(len(data))


class TestDriftDetector:
    def test_distribution_fits(self):
        # Setup
        ref, cols = make_num_data(500)
        test = ref.iloc[np.random.permutation(len(ref))[:10]]
        drift = DriftDetector(**cols)
        drift.fit(ref)

        # Checks
        assert len(drift.check(test)) == 0, "Test data found inconsistent"
        assert len(drift.check(ref.max() + 1)) == len(
            ref.columns
        ), "Maxima not detected"
        assert len(drift.check(ref.min() - 1)) == len(
            ref.columns
        ), "Minima not detected"

    def test_categorical(self):
        df, cols = make_cat_data(10, list("abcd"))
        drift = DriftDetector(**cols)
        drift.fit(df)
        for col in df.columns:
            assert col in drift.bins, f"Column '{col}' rejected"
            assert drift.bins[col] == df[col].value_counts().to_dict()

    def test_add_bins(self):
        yp, _ = make_num_data(100, "randint::0::2")
        df, cols = make_cat_data(10, list("abcd"))
        drift = DriftDetector(**cols)
        drift.fit(df)

        # Test empty
        assert drift.add_bins({}, df)
        assert drift.add_output_bins((), yp)

        # Test actual adding
        new_bins = drift.add_bins(drift.bins, df)
        for col in df.columns:
            assert col in new_bins, f"Column '{col}' rejected"
            assert new_bins[col] == {
                key: 2 * value
                for key, value in df[col].value_counts().to_dict().items()
            }

    def test_storable(self):
        df, cols = make_data(10, cat_choices=list("abc"), num_dists="norm")
        drift = DriftDetector(**cols)
        drift.fit(df)
        json.dumps(drift.bins)
        json.dumps(drift.add_bins(drift.bins, df))
        pred = np.random.randint(0, 2, (100,))
        old = drift.add_output_bins((), pred)
        drift.add_output_bins(old, pred)

    def test_no_drift_warning(self):
        """
        Ensure that minor changes in data do not trigger warnings.
        """
        # Create dummy data
        data_1, cols_1 = make_data(
            500, num_dists=["uniform", "norm"], cat_choices=[list("abc"), list("abc")]
        )
        data_2, cols_2 = make_data(
            10,
            num_dists=[scipy.stats.gamma(1), scipy.stats.beta(1, 2)],
            cat_choices=[list("abc"), list("xyz")],
        )

        # Create dummy predictors
        dummy_model_1 = DummyPredictor()
        dummy_model_2 = DummyPredictor(scipy.stats.randint(0, 10))

        # Instantiate and fit drift detector
        drift = DriftDetector(**cols_1, n_bins=10)
        drift.fit(data_1)
        drift.fit_output(dummy_model_1, data_1)

        # Assert that no DataDriftWarning occurs when given same data and model
        with warnings.catch_warnings(record=True) as caught_warnings:
            # Check drift on input
            drift.check(data_1)
            # Check drift on output
            drift.check_output(dummy_model_1, data_1)
        if any(
            issubclass(warning.category, DataDriftWarning)
            for warning in caught_warnings
        ):
            raise AssertionError("Unnecessary DataDriftWarning detected")

        # Assert that DataDriftWarning occurs when given new data
        with pytest.warns(DataDriftWarning):
            drift.check(data_2)

        # Assert that DataDriftWarning occurs when given new model
        with pytest.warns(DataDriftWarning):
            drift.check_output(dummy_model_2, data_2)



#  Copyright (c) 2022 by Amplo.

import os

import numpy as np
import pandas as pd
import pytest

from amplo import Pipeline
from tests import get_all_modeller_models


class TestPipeline:
    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_main_predictors(self, mode, make_x_y):
        # Test mode
        x, y = make_x_y
        x = x.iloc[:, :5]  # for speed up
        pipeline = Pipeline(n_grid_searches=0, plot_eda=False, extract_features=False)
        pipeline.fit(x, y)
        x_c, _ = pipeline.convert_data(x)

        models = get_all_modeller_models(mode)
        for model in models:
            model.fit(x_c, y)
            pipeline.best_model = model
            pipeline.predict(x)
            assert isinstance(
                pipeline._main_predictors, dict
            ), "Main predictors not dictionary."

    @pytest.mark.parametrize("mode", ["classification"])
    def test_no_dirs(self, mode, make_x_y):
        x, y = make_x_y
        pipeline = Pipeline(no_dirs=True, n_grid_searches=0, extract_features=False)
        pipeline.fit(x, y)
        assert not os.path.exists("Auto_ML"), "Directory created"

    @pytest.mark.parametrize("mode", ["regression"])
    def test_no_args(self, mode, make_x_y):
        x, y = make_x_y
        pipeline = Pipeline(n_grid_searches=0)
        pipeline.fit(x, y)

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_mode_detector(self, mode, make_x_y):
        x, y = make_x_y
        pipeline = Pipeline()
        pipeline._read_data(x, y)._mode_detector()
        assert pipeline.mode == mode

    @pytest.mark.parametrize("mode", ["classification"])
    def test_create_folders(self, mode, make_x_y):
        x, y = make_x_y
        pipeline = Pipeline(n_grid_searches=0)
        pipeline.fit(x, y)

        # Test Directories
        assert os.path.exists("Auto_ML")
        assert os.path.exists("Auto_ML/Data")
        assert os.path.exists("Auto_ML/Production")
        assert os.path.exists("Auto_ML/Results.csv")

    @pytest.mark.parametrize("mode", ["classification"])
    def test_data_organisation(self, mode, make_x_y):
        x, y = make_x_y
        pipeline = Pipeline(n_grid_searches=0)
        pipeline._read_data(x, y)
        pipeline._mode_detector()
        pipeline._data_processing()
        assert os.path.exists("Auto_ML/Data/Cleaned_v1.parquet")
        pipeline._feature_processing()
        assert not os.path.exists("Auto_ML/Data/Cleaned_v1.parquet")
        assert os.path.exists("Auto_ML/Data/Extracted_v1.parquet")
        pipeline._initial_modelling()
        pipeline.conclude_fitting()
        assert len(os.listdir("Auto_ML/Production/v1/")) > 0
        pipeline = Pipeline(n_grid_searches=0)
        assert pipeline.version == 2
        pipeline._read_data(x, y)
        pipeline._mode_detector()
        pipeline._data_processing()
        assert os.path.exists("Auto_ML/Data/Cleaned_v2.parquet")
        assert not os.path.exists("Auto_ML/Data/Extracted_v1.parquet")

    def test_read_write_df(self):
        """
        Check whether intermediate data is stored and read correctly
        """
        # Set path
        data_path = "test_data.parquet"

        # Test single index
        data_write = pd.DataFrame(
            np.random.randint(0, 100, size=(10, 10)),
            columns=[f"feature_{i}" for i in range(10)],
            dtype="int64",
        )
        data_write.index.name = "index"
        Pipeline()._write_df(data_write, data_path)
        data_read = Pipeline._read_df(data_path)
        assert data_write.equals(
            data_read
        ), "Read data should be equal to original data"

        # Test multi-index (cf. IntervalAnalyser)
        data_write = data_write.set_index(data_write.columns[-2:].to_list())
        data_write.index.names = ["log", "index"]
        Pipeline()._write_df(data_write, data_path)
        data_read = Pipeline._read_df(data_path)
        assert data_write.equals(
            data_read
        ), "Read data should be equal to original data"

        # Remove data
        os.remove(data_path)

    @pytest.mark.parametrize("mode", ["classification"])
    def test_capital_target(self, mode, make_x_y):
        x, y = make_x_y
        df = pd.DataFrame(x)
        df["TARGET"] = y
        pipeline = Pipeline(target="TARGET", n_grid_searches=0, extract_features=False)
        pipeline.fit(df)



#  Copyright (c) 2022 by Amplo.

"""
Observer for checking production readiness of model.

This part of code is strongly inspired by [1].

References
----------
[1] E. Breck, C. Shanging, E. Nielsen, M. Salib, D. Sculley (2017).
The ML test score: A rubric for ML production readiness and technical debt
reduction. 1123-1132. 10.1109/BigData.2017.8258038.
"""
from copy import deepcopy

import numpy as np
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.neighbors import KernelDensity

from amplo.classification import PartialBoostingClassifier
from amplo.observation._base import PipelineObserver, _report_obs
from amplo.regression import PartialBoostingRegressor
from amplo.utils.sys import getsize

__all__ = ["ModelObserver"]


class ModelObserver(PipelineObserver):
    """
    Model observer before putting to production.

    While the field of software engineering has developed a full range of best
    practices for developing reliable software systems, similar best-practices
    for ML model development are still emerging.

    The following tests are included:
        1. TODO: Model specs are reviewed and submitted.
        2. TODO: Offline and online metrics correlate.
        3. TODO: All hyperparameters have been tuned.
        4. TODO: The impact of model staleness is known.
        5. A simpler model is not better.
        6. TODO: Model quality is sufficient on important data slices.
        7. TODO: The model is tested for considerations of inclusion.
    """

    _obs_type = "model_observer"

    def observe(self):
        self.check_model_size()
        self.check_better_than_linear()
        self.check_noise_invariance()
        self.check_slice_invariance()
        self.check_boosting_overfit()

    @_report_obs
    def check_model_size(self, threshold=20e6):
        """
        Check the RAM of the model. If it's bigger than 20MB, something is wrong.

        Parameters
        ----------
        threshold : float
            Threshold for latency (in s).

        Returns
        -------
        status_ok : bool
            Observation status. Indicates whether a warning should be raised.
        message : str
            A brief description of the observation and its results.
        """
        ram = getsize(self.model)

        status_ok = ram < threshold
        message = (
            f"A model should occupy more than {threshold // 1e6:.2f}MB of RAM. "
            f"However, the model has a size of {ram // 1e6:.2f}MB."
        )
        return status_ok, message

    @_report_obs
    def check_better_than_linear(self):
        """
        Checks whether the model exceeds a linear model.

        This test incorporates the test ``Model 5`` from [1].

        Citation:
            A simpler model is not better: Regularly testing against a very
            simple baseline model, such as a linear model with very few
            features, is an effective strategy both for confirming the
            functionality of the larger pipeline and for helping to assess the
            cost to benefit tradeoffs of more sophisticated techniques.

        Returns
        -------
        status_ok : bool
            Observation status. Indicates whether a warning should be raised.
        message : str
            A brief description of the observation and its results.
        """
        # Make score for linear model
        if self.mode == self.CLASSIFICATION:
            linear_model = LogisticRegression()
        elif self.mode == self.REGRESSION:
            linear_model = LinearRegression()
        else:
            raise AssertionError("Invalid mode detected.")
        linear_model.fit(self.xt, self.yt)
        linear_model_score = self.scorer(linear_model, self.xv, self.yv)

        # Make score for model to observe
        obs_model_score = self.scorer(self.fitted_model, self.xv, self.yv)

        status_ok = obs_model_score > linear_model_score
        message = (
            "Performance of a linear model should not exceed the "
            "performance of the model to observe. "
            f"Score for linear model: {linear_model_score:.4f}. "
            f"Score for observed model: {obs_model_score:.4f}."
        )
        return status_ok, message

    @_report_obs
    def check_noise_invariance(self):
        """
        This checks whether the model performance is invariant to noise in the data.

        Noise is injected in a slice of the data. The noise follows
        the distribution of the original data.
        Next, the performance metrics are re-evaluated on this noisy slice.

        Returns
        -------
        status_ok : bool
            Observation status. Indicates whether a warning should be raised.
        message : str
            A brief description of the observation and its results.
        """
        # Train model
        model = self.model
        model.fit(self.x, self.y)

        # Inject noise
        signal_noise_ratio = 20
        xn = deepcopy(self.xv)
        for key in self.xv.keys():
            signal_energy = sum(self.xv[key] ** 2)
            noise = np.random.normal(0, 1, len(xn))
            noise_energy = sum(noise**2)
            xn[key] = (
                self.xv[key]
                + np.sqrt(signal_energy / noise_energy * signal_noise_ratio) * noise
            )

        # Arrange message
        status_ok = True
        message = (
            "Model performance deteriorates with realistic noise injection."
            "This indicates too little variance in your data. "
            "Please upload more varied data."
        )

        # Compare performance
        baseline = self._pipe.scorer(model, self.xv, self.yv)
        comparison = self._pipe.scorer(model, xn, self.yv)
        if comparison / baseline < 0.9 or comparison / baseline > 1.1:
            status_ok = False

        return status_ok, message

    @_report_obs
    def check_slice_invariance(self):
        """
        Model performance should be invariant to data slicing.

        Using High Density Regions [1], the weakest slice of 10% data is identified.
        If the optimization metric is significantly (>5%) worse than the average
        metric, a warning is given.

        [1] https://stats.stackexchange.com/questions/148439/what-is-a-highest-density-region-hdr # noqa

        Returns
        -------
        status_ok : bool
            Observation status. Indicates whether a warning should be raised.
        message : str
            A brief description of the observation and its results.
        """
        # Arrange message
        status_ok = True
        message = (
            "Model performs significantly worse on bad slice of the data. "
            "This indicates too little variance in your data. "
            "Please upload more varied data."
        )

        # Normalize
        x = deepcopy(self.x)
        x = (x - x.mean()) / x.std()

        # Fit Kernel Density Estimation & get probabilities
        log_probabilities = (
            KernelDensity(kernel="gaussian", bandwidth=1).fit(x).score_samples(x)
        )
        probabilities = np.exp(log_probabilities)

        # Select smallest slice (10%) (selects per class to avoid imbalance)
        if self.mode == "classification":
            slice_indices = []
            for yc in self.y.unique():
                yc_ind = np.where(self.y == yc)[0]
                samples = int(np.ceil(len(yc_ind) // 10))  # Ceils (to avoid 0)
                slice_indices.extend(
                    yc_ind[np.argpartition(probabilities[yc_ind], samples)[:samples]]
                )
        else:
            slice_indices = np.argpartition(probabilities, int(np.ceil(len(x) // 10)))[
                : int(np.ceil(len(x) // 10))
            ]
        train_indices = [i for i in range(len(x)) if i not in slice_indices]
        xt, xv = self.x.iloc[train_indices], self.x.iloc[slice_indices]
        yt, yv = self.y.iloc[train_indices], self.y.iloc[slice_indices]

        # Train and check performance
        model = self.model
        model.fit(xt, yt)
        score = self._pipe.scorer(model, xv, yv)
        if score < self._pipe.best_score:
            status_ok = False

        return status_ok, message

    @_report_obs
    def check_boosting_overfit(self):
        """
        Checks whether boosting models are overfitted.

        Boosting models are often optimal. Though naturally robust against
        overfitting, it's not impossible to add too many estimators in a
        boosting model, creating complexity to an extent of overfitting.
        This function runs the same model while limiting the estimators, and
        checks if the validation performance decreases.

        Returns
        -------
        status_ok : bool
            Observation status. Indicates whether a warning should be raised.
        message : str
            A brief description of the observation and its results.
        """
        # Check if a boosting model has been selected
        if type(self.model).__name__ not in [
            *PartialBoostingClassifier._SUPPORTED_MODELS,
            *PartialBoostingRegressor._SUPPORTED_MODELS,
        ]:
            return True, ""

        if self.mode == "classification":
            PartialBooster = PartialBoostingClassifier
        else:
            PartialBooster = PartialBoostingRegressor

        # Determine steps & initiate results
        steps = np.ceil(np.linspace(0, PartialBooster.n_estimators(self.model), 7))[
            1:-1
        ]
        scores = []
        for step in steps:
            # Can directly use scorer, as no training is involved at all
            booster = PartialBooster(self.model, step)
            booster._is_fitted = True
            scores.append(self.scorer(booster, self.xv, self.yv))

        # Now, the check fails if there has been a decrease in performance
        status_ok = all(np.diff(scores) / np.max(np.abs(scores)) > 0.001)
        message = (
            "Boosting overfit detected. Please retrain with less estimators. "
            f"Estimators: {steps}, Scores: {scores}"
        )
        return status_ok, message



#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

import warnings

import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import LabelEncoder

from amplo.utils.util import check_dtypes, clean_column_names, clean_feature_name

__all__ = ["clean_feature_name", "DataProcessor"]


class DataProcessor:
    def __init__(
        self,
        target: str | None = None,
        num_cols: list[str] | None = None,
        bool_cols: list[str] | None = None,
        date_cols: list[str] | None = None,
        cat_cols: list[str] | None = None,
        int_cols: list[str] | None = None,  # deprecated
        float_cols: list[str] | None = None,  # deprecated
        include_output: bool = True,
        drop_datetime: bool = False,
        remove_constants: bool = True,
        missing_values: str = "interpolate",
        outlier_removal: str = "clip",
        z_score_threshold: int = 4,
        version: int = 1,
        verbosity: int = 1,
    ):
        """
        Preprocessing Class. Cleans a dataset into a workable format.
        Deals with Outliers, Missing Values, duplicate rows, data types (floats,
        categorical and dates), Not a Numbers, Infinities.

        Parameters
        ----------
        target : str
            Column name of target variable
        num_cols : list
            Float columns
        bool_cols : list
            Boolean columns
        date_cols : list
            Date columns, all parsed to pd.datetime format
        cat_cols : list
            Categorical Columns. Currently, all one-hot encoded.
        int_cols : list[str]
            Deprecated
        float_cols : list[str]
            Deprecated
        include_output : bool
            Whether to include output in the data
        drop_datetime : bool
            Whether to drop datetime columns
        remove_constants : bool
            If False, does not remove constant columns
        missing_values : {"remove_rows", "remove_cols", "interpolate", "mean", "zero"}
            How to deal with missing values.
        outlier_removal : {"quantiles", "z-score", "clip", "none"}
            How to deal with outliers.
        z_score_threshold : int
            If outlier_removal="z-score", the threshold is adaptable
        version : int
            Versioning the output files
        verbosity : int
            How much to print
        """

        # Type checks
        check_dtypes(
            ("target", target, (type(None), str)),
            ("num_cols", num_cols, (type(None), list)),
            ("bool_cols", bool_cols, (type(None), list)),
            ("date_cols", date_cols, (type(None), list)),
            ("cat_cols", cat_cols, (type(None), list)),
            # ignore "int_cols" and "float_cols" as they're deprecated and unused
            ("include_output", include_output, bool),
            # ignore "missing_values" and "outlier_removal" as they're checked below
            ("z_score_threshold", z_score_threshold, int),
            ("remove_constants", remove_constants, bool),
            ("version", version, int),
            ("verbosity", verbosity, int),
        )

        # Integrity checks
        mis_values_algo = ["remove_rows", "remove_cols", "interpolate", "mean", "zero"]
        if missing_values not in mis_values_algo:
            raise ValueError(
                f"Missing values algorithm not implemented, pick from {mis_values_algo}"
            )
        out_rem_algo = ["quantiles", "z-score", "clip", "none"]
        if outlier_removal not in out_rem_algo:
            raise ValueError(
                f"Outlier Removal algorithm not implemented, pick from {out_rem_algo}"
            )

        # Arguments
        self.version = version
        self.include_output = include_output
        self.drop_datetime = drop_datetime
        self.target = target or None
        self.num_cols = num_cols or []
        self.bool_cols = bool_cols or []
        self.cat_cols = cat_cols or []
        self.date_cols = date_cols or []

        # Deprecation warnings
        if int_cols is not None:
            raise DeprecationWarning("int_cols is deprecated.")
        if float_cols is not None:
            raise DeprecationWarning("float_cols is deprecated.")

        # Algorithms
        self.missing_values = missing_values
        self.outlier_removal = outlier_removal
        self.z_score_threshold = z_score_threshold
        self.removeConstants = remove_constants

        # Fitted Settings
        self.data = None
        self.dummies = {}
        self._q1 = None
        self._q3 = None
        self._means = None
        self._stds = None
        self._label_encodings = []

        # Info for Documenting
        self.is_fitted = False
        self.verbosity = verbosity
        self.removed_duplicate_rows = 0
        self.removed_duplicate_columns = 0
        self.removed_outliers = 0
        self.imputed_missing_values = 0
        self.removed_constant_columns = 0

    def _fit_transform(
        self, data: pd.DataFrame, fit=False, remove_constants=True
    ) -> "DataProcessor":
        """
        Wraps behavior of both, fitting and transforming the DataProcessor.
        The function basically reduces duplicated code fragments of `self.fit_transform`
         and `self.transform`.

        Parameters
        ----------
        data : pd.DataFrame
            Input data
        fit : bool
            If True, it will fit the transformer, too
        remove_constants : bool
            If True, it will remove constants when fit

        Returns
        -------
        DataProcessor
        """

        # Clean column names and apply renaming
        if fit:
            self.data, self.rename_dict_ = clean_column_names(data)
            if self.target is not None:
                self.target = self.rename_dict_.get(self.target, None)
            self.num_cols = [self.rename_dict_[col] for col in self.num_cols]
            self.bool_cols = [self.rename_dict_[col] for col in self.bool_cols]
            self.cat_cols = [self.rename_dict_[col] for col in self.cat_cols]
            self.date_cols = [self.rename_dict_[col] for col in self.date_cols]
        else:
            self.data = data.rename(columns=self.rename_dict_)

        # Impute columns
        self._impute_columns()

        # Remove target
        if (
            fit
            and not self.include_output
            and self.target is not None
            and self.target in self.data
        ):
            self.data = self.data.drop(self.target, axis=1)

        if fit:
            # Infer data-types
            self.infer_data_types()

            # Remove Duplicates
            self.remove_duplicates(rows=not isinstance(self.data.index, pd.MultiIndex))

        # Convert data types
        self.convert_data_types(fit_categorical=fit)

        # Remove outliers
        self.remove_outliers(fit=fit)

        # Remove missing values
        self.remove_missing_values()

        # Remove Constants
        if fit and remove_constants:
            self.remove_constants()

        # Convert integer columns
        self.convert_float_int()

        # Encode or decode target
        self._code_target_column(fit=fit)

        return self

    def fit_transform(self, data: pd.DataFrame, remove_constants=True) -> pd.DataFrame:
        """
        Fits this data cleaning module and returns the transformed data.

        Parameters
        ----------
        data : pd.DataFrame
            Input data
        remove_constants : bool
            If True, it will remove constants when fit

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """

        self._fit_transform(data, fit=True, remove_constants=remove_constants)
        assert isinstance(self.data, pd.DataFrame)

        # Finish
        self.is_fitted = True

        return self.data

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Function that takes existing settings (including dummies), and transforms new
        data.

        Parameters
        ----------
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        if not self.is_fitted:
            raise ValueError("Transform only available for fitted objects.")

        self._fit_transform(data, fit=False)
        assert isinstance(self.data, pd.DataFrame)

        return self.data

    def get_settings(self) -> dict:
        """
        Get settings to recreate fitted object.
        """
        assert self.is_fitted, "Object not yet fitted."
        settings = {
            "rename_dict_": self.rename_dict_,
            "num_cols": self.num_cols,
            "bool_cols": self.bool_cols,
            "date_cols": self.date_cols,
            "cat_cols": self.cat_cols,
            "_label_encodings": self._label_encodings,
            "missing_values": self.missing_values,
            "outlier_removal": self.outlier_removal,
            "z_score_threshold": self.z_score_threshold,
            "_means": (
                self._means.to_json() if isinstance(self._means, pd.Series) else None
            ),
            "_stds": (
                self._stds.to_json() if isinstance(self._stds, pd.Series) else None
            ),
            "_q1": self._q1.to_json() if isinstance(self._q1, pd.Series) else None,
            "_q3": self._q3.to_json() if isinstance(self._q3, pd.Series) else None,
            "dummies": self.dummies,
            "fit": {
                "imputed_missing_values": self.imputed_missing_values,
                "removed_outliers": self.removed_outliers,
                "removed_constant_columns": self.removed_constant_columns,
                "removed_duplicate_rows": self.removed_duplicate_rows,
                "removed_duplicate_columns": self.removed_duplicate_columns,
            },
        }
        return settings

    def load_settings(self, settings: dict) -> None:
        """
        Loads settings from dictionary and recreates a fitted object
        """
        self.rename_dict_ = settings.get("rename_dict_", {})
        self.num_cols = settings.get("num_cols", [])
        self.bool_cols = settings.get("bool_cols", [])
        self.date_cols = settings.get("date_cols", [])
        self.cat_cols = settings.get("cat_cols", [])
        self._label_encodings = settings.get("_label_encodings", [])
        self.missing_values = settings.get("missing_values", [])
        self.outlier_removal = settings.get("outlier_removal", [])
        self.z_score_threshold = settings.get("z_score_threshold", [])
        self._means = (
            None
            if settings["_means"] is None
            else pd.read_json(settings["_means"], typ="series")
        )
        self._stds = (
            None
            if settings["_stds"] is None
            else pd.read_json(settings["_stds"], typ="series")
        )
        self._q1 = (
            None
            if settings["_q1"] is None
            else pd.read_json(settings["_q1"], typ="series")
        )
        self._q3 = (
            None
            if settings["_q3"] is None
            else pd.read_json(settings["_q3"], typ="series")
        )
        self.dummies = settings.get("dummies", {})
        self.is_fitted = True

    def infer_data_types(self):
        """
        In case no data types are provided, this function infers the most likely data
        types
        """
        if len(self.cat_cols) == len(self.num_cols) == len(self.date_cols) == 0:
            assert isinstance(self.data, pd.DataFrame)

            # Iterate through keys
            for key in self.data.keys():

                # Skip target
                if key == self.target:
                    continue

                # Remove NaN for feature identification
                f = self.data[key]
                f = f[~f.isna()].infer_objects()

                # Integer and Float
                if pd.api.types.is_integer_dtype(f) or pd.api.types.is_float_dtype(f):
                    self.num_cols.append(key)
                    continue

                # Datetime
                elif pd.api.types.is_datetime64_any_dtype(f):
                    self.date_cols.append(key)
                    continue

                # Booleans
                elif pd.api.types.is_bool_dtype(f):
                    self.bool_cols.append(key)
                    continue

                # Strings / Objects
                elif pd.api.types.is_object_dtype(f):

                    # Check numerical
                    numeric = pd.to_numeric(f, errors="coerce", downcast="integer")
                    if numeric.isna().sum() < 0.3 * len(f):
                        self.num_cols.append(key)

                        # Update data and continue
                        self.data[key] = numeric
                        continue

                    # Check date
                    date = pd.to_datetime(
                        f.astype("str"),
                        errors="coerce",
                        infer_datetime_format=True,
                    )
                    if date.isna().sum() < 0.3 * len(f):
                        self.date_cols.append(key)
                        continue

                    # Check categorical variable
                    if self.data[key].nunique() < max(10, len(self.data) // 4):
                        self.cat_cols.append(key)
                        continue

                # Else not found
                warnings.warn(f"Couldn't identify feature: {key}")

    def convert_data_types(
        self, data=None, fit_categorical: bool = True
    ) -> pd.DataFrame:
        """
        Cleans up the data types of all columns.

        Parameters
        ----------
        data : pd.DataFrame
            Input data
        fit_categorical : bool
            Whether to fit the categorical encoder

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        # Set data
        if data is not None:
            assert isinstance(data, pd.DataFrame)
            self.data = data
        assert isinstance(self.data, pd.DataFrame)

        # Drop Datetime columns
        if self.date_cols and self.drop_datetime:
            warnings.warn(
                f"Data contains datetime columns but are removed: '{self.date_cols}'",
                UserWarning,
            )
            self.data = self.data.drop(self.date_cols, axis=1)
        # Or convert to datetime (before done only on subset)
        elif self.date_cols:
            for key in self.date_cols:
                self.data[key] = pd.to_datetime(
                    self.data[key], errors="coerce", infer_datetime_format=True
                )

        # Integer columns
        for key in self.bool_cols:
            self.data[key] = self.data[key].astype(np.int64)

        # Float columns
        for key in self.num_cols:
            self.data[key] = pd.to_numeric(
                self.data[key], errors="coerce", downcast="float"
            )

        # Categorical columns
        if fit_categorical:
            self._fit_cat_cols()
        else:
            assert self.is_fitted, (
                ".convert_data_types() was called with fit_categorical=False, while "
                "categorical encoder is not yet fitted."
            )
        self.data = self._transform_cat_cols()

        return self.data

    def _fit_cat_cols(self, data=None):
        """
        Encoding categorical variables always needs a scheme. This fits the scheme.
        """
        if data is not None:
            assert isinstance(data, pd.DataFrame)
            self.data = data
        assert isinstance(self.data, pd.DataFrame)

        for key in self.cat_cols:
            # Clean the categorical variables
            is_nan = self.data[key].isna().any()
            self.data[key] = self.data[key].apply(clean_feature_name)

            # Get dummies, convert & store
            dummies = pd.get_dummies(self.data[key], prefix=key, dummy_na=is_nan)
            self.dummies[key] = dummies.keys().tolist()

        return self

    def _transform_cat_cols(self, data=None) -> pd.DataFrame:
        """
        Converts categorical variables according to fitted scheme.
        """
        if data is not None:
            assert isinstance(data, pd.DataFrame)
            self.data = data
        assert isinstance(self.data, pd.DataFrame)

        for key in self.cat_cols:
            # Clean column
            self.data[key] = self.data[key].apply(clean_feature_name)

            # Transform
            value = self.dummies[key]
            dummies = [i[len(key) + 1 :] for i in value]
            str_values = self.data[key].astype("str").values
            self.data[value] = np.equal.outer(str_values, dummies).astype(np.int64)  # type: ignore
            self.data = self.data.drop(key, axis=1)
        return self.data

    def remove_duplicates(self, data=None, rows: bool = True) -> pd.DataFrame:
        """
        Removes duplicate columns and rows.

        Parameters
        ----------
        rows : bool
            Whether to remove duplicate rows --> not desirable to maintain timelines
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        if data is not None:
            assert isinstance(data, pd.DataFrame)
            self.data = data
        assert isinstance(self.data, pd.DataFrame)

        # Note down
        n_rows, n_columns = len(self.data), len(self.data.keys())

        # Remove Duplicates
        if rows:
            self.data = self.data.drop_duplicates()
        self.data = self.data.loc[:, ~self.data.columns.duplicated()]  # type: ignore

        # Note
        self.removed_duplicate_columns = n_columns - len(self.data.keys())
        self.removed_duplicate_rows = n_rows - len(self.data)

        assert isinstance(self.data, pd.DataFrame)
        return self.data

    def remove_constants(self, data=None) -> pd.DataFrame:
        """
        Removes constant columns
        """
        if data is not None:
            assert isinstance(data, pd.DataFrame)
            self.data = data
        assert isinstance(self.data, pd.DataFrame)
        columns = len(self.data.keys())

        # Remove Constants
        if self.removeConstants:
            const_cols = [
                col
                for col in self.data
                if self.data[col].nunique() == 1 and col != self.target
            ]
            self.data = self.data.drop(columns=const_cols)

        # Note
        self.removed_constant_columns = columns - len(self.data.keys())

        return self.data

    def fit_outliers(self, data: pd.DataFrame | None = None):
        """
        Checks outliers
        """
        if data is not None:
            assert isinstance(data, pd.DataFrame)
            self.data = data
        assert isinstance(self.data, pd.DataFrame)

        # With quantiles
        if self.outlier_removal == "quantiles":
            self._q1 = self.data[self.num_cols].quantile(0.25)
            self._q3 = self.data[self.num_cols].quantile(0.75)

        # By z-score
        elif self.outlier_removal == "z-score":
            self._means = self.data[self.num_cols].mean(skipna=True, numeric_only=True)
            self._stds = self.data[self.num_cols].std(skipna=True, numeric_only=True)
            self._stds[self._stds == 0] = 1

    def remove_outliers(
        self, data: pd.DataFrame | None = None, fit: bool = True
    ) -> pd.DataFrame:
        """
        Removes outliers
        """
        if data is not None:
            assert isinstance(data, pd.DataFrame)
            self.data = data
        assert isinstance(self.data, pd.DataFrame)

        # Check if needs fitting
        if fit:
            self.fit_outliers(self.data)
        else:
            if not self.is_fitted:
                raise ValueError(
                    ".remove_outliers() is called with fit=False, yet the object isn't"
                    " fitted yet."
                )

        # With Quantiles
        if self.outlier_removal == "quantiles":
            self.removed_outliers = (
                (self.data[self.num_cols] > self._q3).sum().sum()
                + (self.data[self.num_cols] < self._q1).sum().sum()
            ).tolist()
            self.data[self.num_cols] = self.data[self.num_cols].mask(
                self.data[self.num_cols] < self._q1
            )
            self.data[self.num_cols] = self.data[self.num_cols].mask(
                self.data[self.num_cols] > self._q3
            )

        # With z-score
        elif self.outlier_removal == "z-score":
            z_score = abs((self.data[self.num_cols] - self._means) / self._stds)
            self.removed_outliers = (
                (z_score > self.z_score_threshold).sum().sum().tolist()  # type: ignore
            )
            self.data[self.num_cols] = self.data[self.num_cols].mask(
                z_score > self.z_score_threshold  # type: ignore
            )

        # With clipping
        elif self.outlier_removal == "clip":
            self.removed_outliers = (
                (self.data[self.num_cols] > 1e12).sum().sum()
                + (self.data[self.num_cols] < -1e12).sum().sum()
            ).tolist()
            self.data[self.num_cols] = self.data[self.num_cols].clip(
                lower=-1e12, upper=1e12
            )
        return self.data

    def remove_missing_values(self, data: pd.DataFrame | None = None) -> pd.DataFrame:
        """
        Fills missing values (infinities and "not a number"s)
        """
        if data is not None:
            assert isinstance(data, pd.DataFrame)
            self.data = data
        assert isinstance(self.data, pd.DataFrame)

        # Replace infinities
        self.data = self.data.replace([np.inf, -np.inf], np.nan)

        # Note
        self.imputed_missing_values = (
            self.data[self.num_cols].isna().sum().sum().tolist()
        )

        # Removes all rows with missing values
        if self.missing_values == "remove_rows":
            self.data.dropna(axis=0, inplace=True)

        # Removes all columns with missing values
        elif self.missing_values == "remove_cols":
            self.data.dropna(axis=1, inplace=True)

        # Fills all missing values with zero
        elif self.missing_values == "zero":
            self.data = self.data.fillna(0)

        # Mean and Interpolate require more than 1 value, use zero if less
        elif self.missing_values in ("interpolate", "mean") and len(self.data) <= 1:
            self.data = self.data.fillna(0)

        # Linearly interpolates missing values
        elif self.missing_values == "interpolate":
            # Get all non-date_cols & interpolate
            non_date = np.setdiff1d(self.data.keys().to_list(), self.date_cols)
            self.data[non_date] = self.data[non_date].interpolate(
                limit_direction="both"
            )

            # Fill rest (date & more missing values cols)
            self._interpolate_dates()

        # Fill missing values with column mean
        elif self.missing_values == "mean":
            non_date = np.setdiff1d(self.data.keys().to_list(), self.date_cols)
            self.data[non_date] = self.data[non_date].fillna(self.data.mean())

            # Fill dates
            self._interpolate_dates()

        assert isinstance(self.data, pd.DataFrame)
        return self.data

    def _interpolate_dates(self):
        """
        Unfortunately pandas does not support this out of the box. PR was made, but
        closed pre-merged. https://github.com/pandas-dev/pandas/pull/21915
        """
        assert isinstance(self.data, pd.DataFrame)
        for key in self.date_cols:
            if self.data[key].isna().any():
                unix = self.data[key].astype("int64")
                unix[unix < 0] = np.nan  # NaT are -9e10
                unix = unix.interpolate(method="bfill").interpolate("pad")
                self.data[key] = pd.to_datetime(unix, unit="ns")

    def convert_float_int(self, data=None) -> pd.DataFrame:
        """
        Integer columns with NaN in them are interpreted as floats.
        In the beginning we check whether some columns should be integers,
        but we rely on different algorithms to take care of the NaN.
        Therefore, we only convert floats to integers at the very end
        """
        if data is not None:
            assert isinstance(data, pd.DataFrame)
            self.data = data
        assert isinstance(self.data, pd.DataFrame)

        for key in self.bool_cols:
            if key in self.data:
                self.data[key] = pd.to_numeric(
                    self.data[key], errors="coerce", downcast="integer"
                )
        return self.data

    def _code_target_column(self, fit):
        """En- or decodes target column of `self.data`

        Parameters
        ----------
        fit : bool
            Whether to fit the encoder
        """
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(f"Expected DataFrame but got '{type(self.data)}'.")
        if not self.target or self.target not in self.data:
            return

        # Get labels and encode / decode
        labels = self.data[self.target]  # type: ignore

        # Encode
        self.data[self.target] = self.encode_labels(
            labels, fit=fit, warn_unencodable=False
        )

    def encode_labels(self, labels, *, fit=True, warn_unencodable=True):
        """Encode labels to numerical dtype

        Parameters
        ----------
        labels : np.ndarray or pd.Series
            Labels to encode
        fit : bool
            Whether to (re)fit the label encoder
        warn_unencodable : bool
            Whether to warn when labels are assumed being for regression task

        Returns
        -------
        np.ndarray
            Encoded labels or original labels if unencodable

        Raises
        ------
        NotFittedError
            When no label encoder has yet been trained, i.e. `self._label_encodings` is
            empty
        """
        # Convert to pd.Series for convenience
        labels = pd.Series(labels)

        # It's probably a classification task
        if labels.dtype == object or labels.nunique() <= labels.size / 2:
            # Create encoding
            encoder = LabelEncoder()
            if fit is True:
                # Fit
                encoder.fit(labels)
                self._label_encodings = pd.Series(encoder.classes_).to_list()
            elif not self._label_encodings:
                raise NotFittedError("Encoder it not yet fitted")
            else:
                encoder.fit(self._label_encodings)
            # Encode
            return encoder.transform(labels)

        # It's probably a regression task, thus no encoding needed
        if warn_unencodable:
            warnings.warn(
                UserWarning(
                    "Labels are probably for regression. No encoding happened..."
                )
            )
        return labels.to_numpy()

    def decode_labels(self, labels, *, except_not_fitted=True):
        """Decode labels from numerical dtype to original value

        Parameters
        ----------
        labels : np.ndarray or pd.Series
            Labels to decode
        except_not_fitted : bool
            Whether to raise an exception when label encoder is not fitted

        Returns
        -------
        np.ndarray
            Decoded labels or original labels if label encoder is not fitted and
            `except_not_fitted` is True

        Raises
        ------
        NotFittedError
            When `except_not_fitted` is True and label encoder is not fitted
        """
        try:
            if len(self._label_encodings) == 0:
                raise NotFittedError(
                    "Encoder it not yet fitted. Try first calling `encode_target` "
                    "to set an encoding"
                )
            encoder = LabelEncoder()
            encoder.classes_ = np.array(self._label_encodings)
            return encoder.inverse_transform(labels)
        except NotFittedError as err:
            if except_not_fitted:
                raise err
            else:
                return labels.to_numpy() if isinstance(labels, pd.Series) else labels

    def _impute_columns(self, data: pd.DataFrame | None = None) -> pd.DataFrame:
        """
        *** For production ***
        If a dataset is missing certain columns, this function looks at all registered
        columns and fills them with
        zeros.
        """
        if data is not None:
            assert isinstance(data, pd.DataFrame)
            self.data = data
        assert isinstance(self.data, pd.DataFrame)

        imputed = []
        for keys in [self.num_cols, self.cat_cols]:  # ignoring `self.date_cols`
            for key in [k for k in keys if k not in self.data]:
                self.data[key] = np.zeros(len(self.data))
                imputed.append(key)
        if len(imputed) > 0:
            warnings.warn(f"Imputed {len(imputed)} missing columns! {imputed}")
        return self.data

    def prune_features(self, features: list):
        """
        For use with AutoML.Pipeline. We practically never use all features. Yet this
        processor imputes any missing features. This causes redundant operations,
        memory, and warnings. This function prunes the features to avoid that.

        parameters
        ----------
        features : list
            Required features (NOTE: include required features for extracted)
        """
        hash_features = dict([(k, 0) for k in features])
        self.date_cols = [f for f in self.date_cols if f in hash_features]
        self.num_cols = [f for f in self.num_cols if f in hash_features]
        self.bool_cols = [f for f in self.bool_cols if f in hash_features]
        self.cat_cols = [f for f in self.cat_cols if f in hash_features]



#  Copyright (c) 2022 by Amplo.

from inspect import signature

import pytest
from requests import HTTPError

from amplo.api.databricks import DatabricksJobsAPI

DUMMY_JOB_ID = 749905062420360


class TestDatabricksJobsAPI:
    """
    Tests for the class `DatabricksJobsAPI`.
    """

    @classmethod
    def setup_class(cls):
        cls.api = DatabricksJobsAPI.from_os_env()

    def test_request(self):
        # A valid request should return a dictionary
        out = self.api.request("get", "2.1/jobs/list")
        assert isinstance(out, dict)
        # An invalid request should raise an error
        with pytest.raises(HTTPError):
            self.api.request("get", "")

    def test_list_jobs(self):
        expected_out = self.api.request("get", "2.1/jobs/list", {"limit": 1})
        actual_out = self.api.list_jobs(limit=1)
        assert expected_out == actual_out

    def test_list_runs(self):
        expected_out = self.api.request("get", "2.1/jobs/runs/list", {"limit": 1})
        actual_out = self.api.list_runs(limit=1)
        assert expected_out == actual_out

    def test_get_job(self):
        expected_out = self.api.request("get", "2.1/jobs/get", {"job_id": DUMMY_JOB_ID})
        actual_out = self.api.get_job(DUMMY_JOB_ID)
        assert expected_out == actual_out

    def test_get_run(self):
        # Find an existing run
        runs = self.api.list_runs(limit=1).get("runs")
        if len(runs) == 0:
            pytest.skip("There exists no run.")
        run_id = runs[0]["run_id"]

        # Test
        expected_out = self.api.request("get", "2.1/jobs/runs/get", {"run_id": run_id})
        actual_out = self.api.get_run(run_id)
        assert expected_out == actual_out

    def test_run_job(self):
        run_id = self.api.run_job(DUMMY_JOB_ID)["run_id"]
        assert self.api.cancel_run(run_id) == {}


class TestDatabricksTrainOnCloudNotebook:
    """
    Test functionalities that are used for `train_on_cloud.py` notebook.

    References
    ----------
    - https://github.com/Amplo-AG/Databricks/blob/main/workflows/train_on_cloud.py
    """

    def test_pipeline_format(self):
        # Assert that `Pipeline` is importable
        from amplo import Pipeline

        # Assert that `__init__` accepts those keyword arguments and more
        init_params = signature(Pipeline).parameters.keys()
        expected_init_params = ["main_dir", "name"]
        assert set(init_params).issuperset(expected_init_params)
        assert len(init_params) > len(expected_init_params)  # want to pass more kwargs

    def test_upload_models_format(self):
        # Assert that `upload_model` is importable
        from amplo.api.platform import upload_model

        # Assert that `upload_models` is present and accepts those keyword arguments
        func_params = signature(upload_model).parameters.keys()
        expected_func_params = [
            "model_id",
            "model_dir",
            "team",
            "machine",
            "service",
            "issue",
            "version",
            "host",
            "access_token_os",
        ]
        assert set(expected_func_params) == set(func_params)



#  Copyright (c) 2022 by Amplo.

import warnings
from typing import Union

import numpy as np
import pandas as pd
from scipy import stats

from amplo.utils import hist_search


class DataDriftWarning(Warning):
    pass


class DriftDetector:
    # todo add second order pdf fit
    # todo add subsequence drift detection

    def __init__(
        self,
        num_cols: list = None,
        cat_cols: list = None,
        date_cols: list = None,
        n_bins: int = 500,
        sigma: int = 3,
        with_pdf: bool = False,
    ):
        """
        Detects data drift in streamed input data.
        Supports numerical, categorical and datetime variables.
        Due to streamed, we don't check distributions, just bins.
        Categorical simply checks whether it's not a new column
        Datetime simply checks whether the date is recent
        """
        # Copy kwargs
        self.num_cols = [] if num_cols is None else num_cols
        self.cat_cols = [] if cat_cols is None else cat_cols
        self.date_cols = [] if date_cols is None else date_cols
        self.n_bins = n_bins
        self.with_pdf = with_pdf
        self.sigma = sigma

        # Initialize
        self.bins = {}
        for key in self.cat_cols:
            self.bins[key] = {}
        for key in self.num_cols:
            self.bins[key] = (None, None)
        self.output_bins = (None, None)
        self.distributions = {}

    def fit(self, data: pd.DataFrame) -> "DriftDetector":
        """
        Fits the class object
        """
        # Numerical
        self._fit_bins(data)
        self._fit_distributions(data)

        # Categorical

        return self

    def check(self, data: pd.DataFrame):
        """
        Checks a new dataframe for distribution drift.
        """
        violations = []

        violations.extend(self._check_bins(data))
        violations.extend(self._check_distributions(data))

        return violations

    def fit_output(self, model, data: pd.DataFrame):
        """
        Additionally to detecting input drift, we should also detect output drift. When
        the distribution of predicted outcomes change, it's often a sign that some under
        laying dynamics are shifting.
        """
        assert hasattr(model, "predict"), "Model does not have 'predict' attribute."

        # If it's a classifier and has predict_proba, we use that :)
        if hasattr(model, "predict_proba"):
            prediction = model.predict_proba(data)[:, 1]
        else:
            prediction = model.predict(data)

        ma, mi = max(prediction), min(prediction)
        y, x = np.histogram(
            prediction,
            bins=self.n_bins,
            range=(mi - (ma - mi) / 10, ma + (ma - mi) / 10),
        )
        self.output_bins = (x.tolist(), y.tolist())

    def check_output(self, model, data: pd.DataFrame, add: bool = False):
        """
        Checks the predictions of a model.
        """
        assert hasattr(model, "predict"), "Model does not have 'predict' attribute."

        # If it's a classifier and has predict_proba, we use that :)
        if hasattr(model, "predict_proba"):
            prediction = model.predict_proba(data)[:, 1]
        else:
            prediction = model.predict(data)

        # Check all predictions
        x, y = self.output_bins
        count_drifts = 0
        for value in prediction:
            ind = hist_search(x, value)
            if ind == -1 or y[ind] <= 0:
                # Drift detected
                count_drifts += 1
        if count_drifts > 0:
            severity = count_drifts / len(prediction) * 100
            warnings.warn(
                DataDriftWarning(f"Output drift detected! Severity: {severity:.2f}%")
            )

        # Add new output
        if add:
            y += np.histogram(prediction, bins=x)
            return y

    def get_weights(self) -> dict:
        """
        Gets the weights of the fitted object.
        Useful to save :)
        """
        return {
            "bins": self.bins,
            "output_bins": self.output_bins,
            "distributions": self.distributions,
        }

    def load_weights(self, weights):
        """
        Sets the weights of the object to recreate a previously fitted object.

        Parameters
        ----------
        weights : typing.Dict[str, dict or tuple]
            Weights of a (fitted) object.
            Expected keys are:
            - "bins" (dict):
                Bins dictionary with bins and quantities for all numeric keys.
            - "output_bins" (tuple):
                Output bins.
            - "distributions" (dict):
                Fitted distributions for all numeric keys.
        """
        self.bins = weights.get("bins", {})
        self.output_bins = weights.get("output_bins", (None, None))
        self.distributions = weights.get("distributions", {})
        return self

    def _fit_bins(self, data: pd.DataFrame):
        """
        Fits a histogram on each numerical column.
        """
        # Fit numerical
        for key in self.num_cols:
            ma, mi = data[key].max(), data[key].min()
            y, x = np.histogram(
                data[key],
                bins=self.n_bins,
                range=(mi - (ma - mi) / 10, ma + (ma - mi) / 10),
            )
            self.bins[key] = (x.tolist(), y.tolist())

        # Fit categorical
        for key in self.cat_cols:
            self.bins[key] = data[key].value_counts().to_dict()

    def _check_bins(self, data: pd.DataFrame, add: bool = False):
        """
        Checks if the current data falls into bins
        """
        violations = []

        for key in self.num_cols:
            # Get bins
            x, y = self.bins[key]

            # Check bins
            if isinstance(data, pd.DataFrame):
                for value in data[key].values:
                    ind = hist_search(x, value)
                    if ind == -1 or y[ind] <= 0:
                        violations.append(key)
                        break
            elif isinstance(data, pd.Series):
                ind = hist_search(x, data[key])
                if ind == -1 or (
                    y[ind] <= 0
                    and y[max(0, ind - 1)] <= 0
                    and y[min(self.n_bins, ind + 1)] <= 0
                ):
                    violations.append(key)

            # Add data
            if add:
                y += np.histogram(data[key], bins=x)
                self.bins[key] = (x, y)

        if len(violations) > 0:
            warnings.warn(
                DataDriftWarning(
                    f"Drift detected! "
                    f"{len(violations)} features outside training bins: {violations}"
                )
            )

        return violations

    def _fit_distributions(self, data: pd.DataFrame):
        """
        Fits a distribution on each numerical column.
        """
        if self.with_pdf:
            distributions = ["gamma", "beta", "dweibull", "dgamma"]
            distances = []
            fitted = []

            # Iterate through numerical columns
            for key in self.num_cols:
                y, x = np.histogram(data[key], normed=True)
                x = (x + np.roll(x, -1))[:-1] / 2.0  # Get bin means

                # Iterate through distributions
                for distribution in distributions:
                    # Fit & Get PDF
                    dist = getattr(stats, distribution)

                    # Multiple order fit
                    params = dist.fit(data[key])
                    fitted_pdf = dist.pdf(
                        x, loc=params[-2], scale=params[-1], *params[:-2]
                    )

                    # Analyse
                    distances.append(sum((y - fitted_pdf) ** 2))
                    fitted.append(
                        {
                            "distribution": distribution,
                            "params": params,
                        }
                    )

                # Select lowest
                self.distributions[key] = fitted[np.argmin(distances)]

    def _check_distributions(self, data: pd.DataFrame) -> list:
        """
        Checks whether the new data falls within the fitted distributions
        """
        # Init
        violations = []

        if self.with_pdf:
            # Check all numerical columns
            for key in self.num_cols:
                dist = getattr(stats, self.distributions[key]["distribution"])
                params = self.distributions[key]["params"]
                probabilities = dist.pdf(
                    data[key].values, loc=params[-2], scale=params[-1], *params[:-2]
                )

                if any(p < self.sigma for p in probabilities):
                    violations.append(key)
                    continue

            if len(violations) > 0:
                warnings.warn(
                    f"Drift detected! {len(violations)} features outside training bins:"
                    f" {violations}",
                    DataDriftWarning,
                )

        return violations

    def add_output_bins(
        self, old_bins: tuple, prediction: Union[np.ndarray, pd.Series]
    ):
        """
        Just a utility, adds new data to an old distribution.
        """
        if len(old_bins) != 0:
            x, y = old_bins
            yn = np.histogram(prediction, bins=x)[0].tolist()
            y = [y[i] + yn[i] for i in range(len(y))]
        else:
            y, x = np.histogram(prediction, bins=self.n_bins)
            x, y = x.tolist(), y.tolist()
        return x, y

    def add_bins(self, bins: dict, data: pd.DataFrame):
        """
        Just a utility, adds new data to an old distribution.
        """
        # Add numerical bins
        for key in self.num_cols:
            if key in data:
                if key in bins:
                    x, y = bins[key]
                elif key in self.bins:
                    x, _ = self.bins[key]
                    y = [0 for i in range(len(x) - 1)]
                else:
                    raise ValueError(f"Drift Detector - Adding unfitted feature. {key}")
                yn = np.histogram(data[key], bins=x)[0].tolist()
                y = [y[i] + yn[i] for i in range(len(y))]
                bins[key] = (x, y)

        # Add categorical
        for key in self.cat_cols:
            if key in data:
                counts = data[key].value_counts().to_dict()
                bins[key] = {
                    k: counts.get(k, 0) + bins.get(key, {}).get(k, 0)
                    for k in counts.keys() | bins.get(key, {}).keys()
                }
        return bins



#  Copyright (c) 2022 by Amplo.

import itertools
import random

import pytest

from amplo.api import API
from amplo.utils.testing import make_interval_data, make_production_data
from tests.unit.api import TestAPI


class TestInterface(TestAPI):
    def test_iterate_directory_local(self):
        api = API(self.sync_dir, verbose=2)

        # Make local dummy folders
        dummy_subdirs = [
            (self.sync_dir / dummy_subdir).resolve()
            for dummy_subdir in ("test1", "test2", "third_folder")
        ]
        for folder in dummy_subdirs:
            (self.sync_dir / folder).mkdir(parents=True)

        # Check yielding `all` subdirectories
        for selection in (None, "*", ".*"):
            all_subdirs = list(
                api._iterate_directory(
                    parent_dir=self.sync_dir, selection=selection, from_local=True
                )
            )
            assert set(dummy_subdirs) == set(all_subdirs), (
                f"Setting ``selection={selection}`` should return all "
                "subdirectories in the dummy folder."
            )

        # Check yielding an explicit selection
        test_subdirs = list(
            api._iterate_directory(
                parent_dir=self.sync_dir, selection="test1", from_local=True
            )
        )
        assert set(i.resolve() for i in self.sync_dir.glob("test1")) == set(
            test_subdirs
        ), "Invalid explicit selection."
        test_subdirs_via_list = list(
            api._iterate_directory(
                parent_dir=self.sync_dir, selection=["test1"], from_local=True
            )
        )
        assert set(test_subdirs_via_list) == set(
            test_subdirs
        ), "Giving either a str or list of str should yield the same results."

        # Check yielding a selection via RegEx
        test_subdirs = list(
            api._iterate_directory(
                parent_dir=self.sync_dir, selection="test.*", from_local=True
            )
        )
        assert set(i.resolve() for i in self.sync_dir.glob("test*")) == set(
            test_subdirs
        ), "Invalid RegEx selection."

    def test_iterate_directory_blob(self):
        api = API(self.sync_dir, verbose=2)

        all_subdirs = list(
            api._iterate_directory(parent_dir="Demo", selection="*", from_local=False)
        )
        test_subdirs = list(
            api._iterate_directory(
                parent_dir="Demo", selection="Charger 150kW.*", from_local=False
            )
        )
        assert len(test_subdirs) > 0, "List of directories should not be empty."
        assert set(test_subdirs).issubset(
            all_subdirs
        ), "`test` directories should be a subset of `all` directories."
        assert set(test_subdirs) != set(
            all_subdirs
        ), "`test` directories should not be equal to `all` directories."

    def test_iter_data(self):
        # Create dummy dataset
        paths_orig, infos_orig = [], []
        rng = random.Random(354687657938)
        for t, m, s in itertools.product(range(3), repeat=3):
            if rng.choice([True, False]):
                # Create "gaps" in data / leave some out
                continue
            team, machine, service = f"Team{t}", f"Machine{m}", f"Service{s}"
            paths_orig += [(self.sync_dir / f"{team}/{machine}/{service}").resolve()]
            infos_orig += [dict(team=team, machine=machine, service=service)]
            make_interval_data(directory=paths_orig[-1] / "data")

        # Define helper function
        def is_selected(orig_lookup: dict, check_lookup: dict, name: str):
            """Return True if a selection match was found"""
            orig = orig_lookup[f"{name}"]
            check = check_lookup[f"{name}s"]
            if check is None or check == "*" or orig in check:
                return True
            else:
                return False

        # Define test cases
        select_kwargs = [
            # Get all combinations
            dict(teams=None, machines=None, services=None),
            # Single selection
            dict(teams="Team0", machines="Machine0", services="Service0"),
            # Mixed selection
            dict(teams=None, machines="Machine0", services=["Service0", "Service1"]),
        ]
        for kwargs in select_kwargs:
            # Get all paths from ``_iter_data``
            paths_actual, infos_actual = [], []
            for path, team, machine, service in API(
                self.sync_dir, **kwargs
            )._iterate_data(level="service"):
                paths_actual += [path]
                infos_actual += [dict(team=team, machine=machine, service=service)]

            # Define subset
            paths_target, infos_target = [], []
            for path, info in zip(paths_orig, infos_orig):
                if any(not is_selected(info, kwargs, key) for key in info.keys()):
                    continue
                paths_target += [path]
                infos_target += [info]

            # Assertions
            assert set(paths_actual) == set(
                paths_target
            ), f"Path check has failed with kwargs: {kwargs}"
            assert set(str(i) for i in infos_actual) == set(
                str(i) for i in infos_target
            ), f"Info check has failed with kwargs: {kwargs}"

    def test_download(self):
        # TODO: Set up a data selection that is quick to download
        # select = dict(teams='Amplo', machines='TestMachine', services='Unlabelled')
        select = dict(
            teams="Demo",
            machines="Charger 75kW",
            services="Diagnostics",
            issues="Blocked Radiator",
        )
        # Set up API and synch data
        api = API(self.sync_dir, **select, verbose=2)
        api.sync_data()
        # Assert existence of folder
        assert (
            self.sync_dir
            / select["teams"]
            / select["machines"]
            / select["services"]
            / "data"
            / select["issues"]
        ).exists(), "Download has failed or is not in assumed directory"

    def test_training(self):
        # Create a small dummy dataset
        dummy_data_path = (
            self.sync_dir / "DummyTeam" / "DummyMachine" / "DummyService" / "data"
        )
        make_interval_data(n_labels=3, directory=dummy_data_path)
        # Set up api and train models
        api = API(self.sync_dir, verbose=2)
        api.train_models(n_grid_searches=0)
        # Assertions
        assert (
            len(api._trained_model_args) > 0
        ), "No model has been trained or failed to append model arguments to list"

    def test_upload(self):
        """
        This only tests that upload_latest_model doesn't error.
            - Good to check whether the file exists in the cloud.
            - Also needs some cleanup to make sure it doesn't stay there.
            - Also does upload_latest_model truly take the latest?
            - What if latest locally and in the cloud are outdated? It needs to go in
              the right folder.
        """
        # TODO: Implement as soon as platform is out of beta mode
        pytest.skip("Test not yet implemented")

        # Make dummy production data
        issue_dir, kwargs = make_production_data(
            self.sync_dir, team="Demo", machine="Charger 75kW", service="Diagnostics"
        )
        # Set up api and upload model
        api = API(self.sync_dir, verbose=2)
        api.upload_models(
            [
                [
                    issue_dir,
                    kwargs["team"],
                    kwargs["machine"],
                    kwargs["service"],
                    kwargs["issue"],
                    kwargs["version"],
                ]
            ]
        )
